#!/usr/bin/env python

# ATTENTION:
# If you are reading this, you have opened the file in a text editor.
# If you would like to *run* this example rather than *read* it, you
# should open Terminal.app, drag this document's icon onto the terminal
# window, bring Terminal.app to the foreground (if necessary) and hit return.

# Propeller Serial Port Loader (version 0.3)
# Chad George with MacOS changes by Michael Green
# 10-31-06, 12-04-07, 12-07-07

import serial
import time
import string
import sys
import os
import re

def findCocoaDialog():
    possiblePlaces = ["/Applications", "/Library/Application Support",
                      "~/Applications", "~/Library/Application Support"]
    for theBase in possiblePlaces:
        thePath = os.path.expanduser(os.path.join(theBase,"CocoaDialog.app"))
        if os.path.exists(thePath):
            thePath = os.path.join(thePath,"Contents/MacOS/CocoaDialog")
            return "\"" + thePath + "\""
    return ""
    
class ProgressBar:
    """Simple class for displaying progress bars using CocoaDialog"""

    def __init__(self, title="Progress", message="", percent=-1):
        """Create progress bar dialog"""
        if percent < 0:
            template = "%s progressbar --title '%s' --text '%s' --indeterminant"
            template = template % (findCocoaDialog(), title, message)
        else:
            template = "%s progressbar --title '%s' --text '%s' --percent %d"
            template = template % (findCocoaDialog(), title, message, percent)
        self.percent = percent
        self.pipe = os.popen(template, "w")
        self.message = message
            
    def update(self, percent=0, message=False):
        """Update progress bar (and message if desired)"""
        if message:
            self.message = message  # store message for persistence
        self.pipe.write("%d %s\n" % (percent, self.message))
        self.pipe.flush()
        
    def finish(self):
        """Close progress bar window"""
        self.pipe.close()

Comm = serial.Serial()
CommPort = None

TxBuffSize = 256
RxBuffSize = 256

TxBuff = []
RxBuff = []
LFSR = 0
Version = 0
VersionMode = True
LongCount = 0

man_ex_1 = [0x00, 0x1B, 0xB7, 0x00, 0x00, 0xDE, 0x10, 0x00, 0x30, 0x00, 0x38,
            0x00, 0x18, 0x00, 0x3C, 0x00, 0x20, 0x00, 0x02, 0x00, 0x08, 0x00,
            0x00, 0x00, 0x37, 0x03, 0x3D, 0xD6, 0x1C, 0x37, 0x03, 0x3D, 0xD4,
            0x47, 0x3A, 0x2D, 0xC6, 0xC0, 0x3F, 0x91, 0xEC, 0x23, 0x04, 0x71,
            0x32, 0x00, 0x00, 0x00]

# Reset hardware on current comm port
def ResetHardware():
    if Comm.isOpen():
        Comm.setDTR(1)
        DelayMSec(25)
        Comm.setDTR(0)
        DelayMSec(100)
        Comm.flushInput()

# Find hardware on any comm port
def FindAnyHardware():
    global CommPort

    result = "Can't open a comm port"
    
    if OpenComm(CommPort):
        result = HardwareFound()
        if result == True:
            return True

    for port in ValidCommPorts():
        if OpenComm(port):
            result = HardwareFound()
            if result == True:
                return True

    return result

# Check for hardware on current comm port
def HardwareFound():
    global LFSR
    global Version

    if Comm.isOpen():
        ResetHardware()
        TByte(0xF9)
        LFSR = ord('P')
        for i in range(250):
            TByte( IterateLFSR() | 0xFE )
        for i in range(250 + 8):
            TByte(0xF9)
        TComm()

        for i in range(250):
            bit = RBit(False, 100)
            if bit < 0:
                return CommError()
            if bit <> IterateLFSR():
                return "No Propeller Found - Bad response"
                  
        for i in range(8):
            Version = ((Version >> 1) & 0x7F) | (RBit(False, 50) << 7)

        if VersionMode:
            TLong(0)
            TComm()
            ResetHardware()
            CloseComm()

        return True
    return "No Propeller Found - Can't open"

# Get hardware version
def GetHardwareVersion():
    global VersionMode

    VersionMode = True
    return FindAnyHardware()

# Send a buffer of data to the hardware
def TalkToHardware(bar, command, buffer):
    global VersionMode
    global Version
    global CommPort

    bar.update(message="Looking for hardware...")
    VersionMode = False
    err = FindAnyHardware()
    if err <> True:
        if Comm.isOpen():
            Comm.close()
        return err

    bar.update(message="Found Propeller Version %s on %s" % (Version, CommPort))

    TLong(command)
    TComm;

    if command > 0:
        longcount = len(buffer) >> 2
        bar.update(message="Sending program (%s Words)..." % longcount)

        TLong(longcount)

        for x in range(0, longcount << 2, 4):
            long = (ord(buffer[x]) |
                    (ord(buffer[x+1]) << 8)  |
                    (ord(buffer[x+2]) << 16) |
                    (ord(buffer[x+3]) << 24))

            TLong(long)

        TComm()

        bar.update(message="Verifying RAM Checksum...")
        bit = RBit(True, 2500)
        if bit < 0:
            return HardwareLost()
        if bit == 1:
            return CommError("RAM Checksum Error")

        if command > 1:
            bar.update(message="Programming EEPROM...")
            bit = RBit(True, 5000)
            if bit < 0:
                return CommError()
            if bit == 1:
                CommError("EEPROM Programming Error")

            bar.update(message="Verifying EEPROM...")
            bit = RBit(True, 2500)
            if bit < 0:
                return CommError()
            if bit == 1:
                CommError("EEPROM Verify Error")

    CloseComm()
    return "Done."

# Cycles the LFSR
def IterateLFSR():
    global LFSR

    bit = LFSR & 1
    LFSR = ((LFSR << 1) & 0xFE) | (LFSR >> 7 ^
                                   LFSR >> 5 ^
                                   LFSR >> 4 ^
                                   LFSR >> 1) & 1
    return bit

# Communication Routines

# Returns a list of comm ports to try to find hardware on
def ValidCommPorts():
    if sys.platform == "win32":
        return ["COM%d" % x for x in range(9,0,-1)]
    else:
        devices = os.listdir("/dev")
        serial = []
        pattern = re.compile("tty.usbserial\-.*")
        for name in devices:
            if pattern.match(name) <> None:
               serial = serial + ["/dev/" + name]
        return serial

# Open a comm port at 115200 baud, 8-N-1
def OpenComm(port):
    global Comm
    global CommPort

    CloseComm()
    try:
        Comm.port = port
        CommPort = port
        Comm.baudrate = 115200
        Comm.timeout = 0
        Comm.open()                 # open port 115200-8-N-1
    except serial.SerialException:
        return False
    return True

# Close the comm port
def CloseComm():
    global Comm

    if Comm.isOpen():
        Comm.close()

# Handle a communications error
def CommError(msg="Hardware contact lost"):
    global Comm

    if Comm.isOpen():
       Comm.close()

    return msg

# Send the trasmit buffer to the hardware
def TComm():
    global TxBuff
    global Comm

    sbuff = ''.join(map(chr, TxBuff))
    if Comm.isOpen():
        Comm.write(sbuff)
    TxBuff = []

# Add a byte to the transmit buffer
def TByte(x):
    global TxBuff
    global TxBuffSize

    TxBuff.append(x)
    if len(TxBuff) == TxBuffSize:
        TComm()

# Add encoded long (11 bytes) to transmit buffer
def TLong(x):
    global TByte

    for i in range(11):
        TByte(0x92 |
              (-(i==10) & 0x60) |
              (x & 1) |
              ((x & 2) << 2) |
              ((x & 4) << 4) )
        x = x >> 3

# Receive response from hardware via echoed byte or -1 on error
def RBit(echo, timeout):
    global RxBuff
    global default_timer

    timeout = timeout / 1000.0
    start = default_timer()
    while (default_timer() - start) < timeout:
        if echo:
            TByte(0xF9)
            TComm()
            DelayMSec(25)
        if len(RxBuff) == 0:
            RComm()
        else:
            bit = RxBuff[0] - 0xFE
            del RxBuff[0]

            if (bit & 0xFE) == 0:
                return bit

            return -1
    return -1

def RComm():
    global RxBuff
    global Comm

    RxBuff = []
    if Comm.isOpen():
        RxBuff = map(ord, Comm.read(RxBuffSize))


# Miscellaneous Utility Routines

LoopDelayMode = False

if sys.platform == "win32":
    default_timer = time.clock      # higher resolution for windows
else:
    default_timer = time.time       # higher resolution for others

def DelayMSec(ms):
    global LoopDelayMode
    global default_timer

    ms = ms / 1000.0

    if LoopDelayMode:
        start = default_timer()
        while (default_timer() - start) < ms:
            pass
    else:
        time.sleep(ms)

if __name__ == "__main__":
    thePath = findCocoaDialog()
    template = "%s fileselect --string-output --text %s --with-extensions %s"
    message = "Select file to download"
    extensions = "binary bin"
    pipe = os.popen(template % (thePath, message, extensions),"r")
    lines = re.split("\n",pipe.read())
    pipe.close()
    if lines[0] <> "":
        progress = ProgressBar(title="Propeller Download",
                               message="Opening file...")
        f = open(lines[0],"rb")
        b = f.read()
        f.close()
        message = TalkToHardware(progress, 3, b)
        progress.finish()
        template ="%s ok-msgbox --text \"Propeller Download\"" % (thePath)
        template = template + " --informative-text \"%s\" --no-cancel --icon info"
        pipe = os.popen(template % (message),"r")
        lines = pipe.read()
        pipe.close()
